@@ -30,6 +30,7 @@
 #include <sys/socket.h>
 #include <sys/un.h>
 #include <netinet/in.h>
+#include <netinet/tcp.h>
 #ifdef DNETCONN
 #include <netdnet/dnetdb.h>
 #include <netdnet/dn.h>
@@ -125,7 +126,7 @@ static int _xcb_open_unix(char *protocol, const char *file);
 static int _xcb_open_decnet(const char *host, char *protocol, const unsigned short port);
 #endif
 #ifdef HAVE_ABSTRACT_SOCKETS
-static int _xcb_open_abstract(char *protocol, const char *file);
+static int _xcb_open_abstract(char *protocol, const char *file, size_t filelen);
 #endif
 
 static int _xcb_open(char *host, char *protocol, const int display)
@@ -135,6 +136,7 @@ static int _xcb_open(char *host, char *protocol, const int display)
 #endif
     static const char base[] = "/tmp/.X11-unix/X";
     char file[sizeof(base) + 20];
+    int filelen;
 
     if(*host)
     {
@@ -161,9 +163,13 @@ static int _xcb_open(char *host, char *protocol, const int display)
     }
 
     /* display specifies Unix socket */
-    snprintf(file, sizeof(file), "%s%d", base, display);
+    filelen = snprintf(file, sizeof(file), "%s%d", base, display);
+    if(filelen < 0)
+        return -1;
+    /* snprintf may truncate the file */
+    filelen = MIN(filelen, sizeof(file) - 1);
 #ifdef HAVE_ABSTRACT_SOCKETS
-    fd = _xcb_open_abstract(protocol, file);
+    fd = _xcb_open_abstract(protocol, file, filelen);
     if (fd >= 0 || (errno != ENOENT && errno != ECONNREFUSED))
         return fd;
 
@@ -188,8 +194,9 @@ static int _xcb_open_decnet(const char *host, const char *protocol, const unsign
     addr.sdn_add.a_len = nodeaddr->n_length;
     memcpy(addr.sdn_add.a_addr, nodeaddr->n_addr, addr.sdn_add.a_len);
 
-    sprintf((char *)addr.sdn_objname, "X$X%d", port);
-    addr.sdn_objnamel = strlen((char *)addr.sdn_objname);
+    addr.sdn_objnamel = sprintf((char *)addr.sdn_objname, "X$X%d", port);
+    if(addr.sdn_objnamel < 0)
+        return -1;
     addr.sdn_objnum = 0;
 
     fd = socket(PF_DECnet, SOCK_STREAM, 0);
@@ -197,8 +204,9 @@ static int _xcb_open_decnet(const char *host, const char *protocol, const unsign
         return -1;
 
     memset(&accessdata, 0, sizeof(accessdata));
-    sprintf((char*)accessdata.acc_acc, "%d", getuid());
-    accessdata.acc_accl = strlen((char *)accessdata.acc_acc);
+    accessdata.acc_accl = sprintf((char*)accessdata.acc_acc, "%d", getuid());
+    if(accessdata.acc_accl < 0)
+        return -1;
     setsockopt(fd, DNPROTO_NSP, SO_CONACCESS, &accessdata, sizeof(accessdata));
 
     if(connect(fd, (struct sockaddr *) &addr, sizeof(addr)) == -1) {
@@ -250,6 +258,10 @@ static int _xcb_open_tcp(char *host, char *protocol, const unsigned short port)
     {
         fd = socket(addr->ai_family, addr->ai_socktype, addr->ai_protocol);
         if(fd >= 0) {
+            int on = 1;
+            setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &on, sizeof(on));
+	    setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof(on));
+
             if (connect(fd, addr->ai_addr, addr->ai_addrlen) >= 0)
                 break;
             close(fd);
@@ -284,7 +296,7 @@ static int _xcb_open_unix(char *protocol, const char *file)
 }
 
 #ifdef HAVE_ABSTRACT_SOCKETS
-static int _xcb_open_abstract(char *protocol, const char *file)
+static int _xcb_open_abstract(char *protocol, const char *file, size_t filelen)
 {
     int fd;
     struct sockaddr_un addr = {0};
@@ -295,9 +307,9 @@ static int _xcb_open_abstract(char *protocol, const char *file)
 
     strcpy(addr.sun_path + 1, file);
     addr.sun_family = AF_UNIX;
-    namelen = offsetof(struct sockaddr_un, sun_path) + 1 + strlen(file);
+    namelen = offsetof(struct sockaddr_un, sun_path) + 1 + filelen;
 #ifdef HAVE_SOCKADDR_SUN_LEN
-    addr.sun_len = 1 + strlen(file);
+    addr.sun_len = 1 + filelen;
 #endif
     fd = socket(AF_UNIX, SOCK_STREAM, 0);
     if (fd == -1)
@@ -312,39 +324,7 @@ static int _xcb_open_abstract(char *protocol, const char *file)
 
 xcb_connection_t *xcb_connect(const char *displayname, int *screenp)
 {
-    int fd, display = 0;
-    char *host;
-    char *protocol;
-    xcb_connection_t *c;
-    xcb_auth_info_t auth;
-    
-    int parsed = _xcb_parse_display(displayname, &host, &protocol, &display, screenp);
-
-#ifdef HAVE_LAUNCHD
-    if(!displayname)
-        displayname = getenv("DISPLAY");
-    if(displayname && strlen(displayname)>11 && !strncmp(displayname, "/tmp/launch", 11))
-        fd = _xcb_open_unix(NULL, displayname);
-    else
-#endif
-    if(!parsed)
-        return (xcb_connection_t *) &error_connection;
-    else
-        fd = _xcb_open(host, protocol, display);
-    free(host);
-
-    if(fd == -1)
-        return (xcb_connection_t *) &error_connection;
-
-    if(_xcb_get_auth_info(fd, &auth, display))
-    {
-        c = xcb_connect_to_fd(fd, &auth);
-        free(auth.name);
-        free(auth.data);
-    }
-    else
-        c = xcb_connect_to_fd(fd, 0);
-    return c;
+    return xcb_connect_to_display_with_auth_info(displayname, NULL, screenp);
 }
 
 xcb_connection_t *xcb_connect_to_display_with_auth_info(const char *displayname, xcb_auth_info_t *auth, int *screenp)
@@ -352,6 +332,8 @@ xcb_connection_t *xcb_connect_to_display_with_auth_info(const char *displayname,
     int fd, display = 0;
     char *host;
     char *protocol;
+    xcb_auth_info_t ourauth;
+    xcb_connection_t *c;
 
     int parsed = _xcb_parse_display(displayname, &host, &protocol, &display, screenp);
     
@@ -371,5 +353,17 @@ xcb_connection_t *xcb_connect_to_display_with_auth_info(const char *displayname,
     if(fd == -1)
         return (xcb_connection_t *) &error_connection;
 
-    return xcb_connect_to_fd(fd, auth);
+    if(auth)
+        return xcb_connect_to_fd(fd, auth);
+
+    if(_xcb_get_auth_info(fd, &ourauth, display))
+    {
+        c = xcb_connect_to_fd(fd, &ourauth);
+        free(ourauth.name);
+        free(ourauth.data);
+    }
+    else
+        c = xcb_connect_to_fd(fd, 0);
+
+    return c;
 }
